# Copyright 2016 Jochen Fabricius
# Distribution and modification is encouraged

proc :w {{dirname ""} {overwrite no} {create no}} {
  if {[setup cget System,LastDir] eq {}} {
    set initdir [file join [pwd]]
  } else {
    set initdir [setup cget System,LastDir]
  }
  # get all overlays
  if {[llength [::overlay::ids]] <=0} {
    ::cmdbar::writeToCmdBar {Nothing to save} -bg yellow
    return 0
  }
  if {$dirname == {}} {
    if {[set dirname [tk_chooseDirectory \
      -initialdir $initdir \
      -mustexist yes]] == {}} {
      ::cmdbar::writeToCmdBar "Save cancelled" -bg yellow
      return 0
    }
  }
  if {![file isdirectory $dirname]} {
    if {$create} {
      file create $dirname
    } else {
      return -code error "Directory $dirname does not exist"
    }
  }
  setup configure System,LastDir $dirname
  # open index file
  if {[file exists [file join $dirname index]] && !$overwrite} {
    if {[tk_messageBox -message {Index file exists. Overwrite?} -default no -icon question -type yesno] eq {no}} {
      ::cmdbar::writeToCmdBar "Save cancelled" -bg yellow
      return 0
    } else {
      set overwrite {yes}
    }
  }
  if {[set idx [open [file join $dirname index] w]] eq {}} {
    return -code error "Couldn't open index file"
  }
  # write each IR to a file and fill index file
  foreach {id} [::overlay::ids] {
    set tail #${id}_[getProp $id Name]
    set fname [file join $dirname $tail]
    if {[catch {writeFile $id $fname $overwrite} result]} {
      close $idx
      return -code error $result
    }
    foreach {prop val} $result {
      if {($prop eq {Mic}) || ($prop eq {MicExp})} continue
      puts -nonewline $idx "[list $prop $val] "
    }
    # add the filename
    puts -nonewline $idx [list Filename ./$tail]
    puts $idx {}
  }
  close $idx

  # success
  ::cmdbar::writeToCmdBar "Saved to $dirname" -bg green
  return 0
}

# internal function, no validity checks
proc writeFile {idorObj {fname {}} {overwrite no}} {
  if {[file exists $fname] && !$overwrite} {
    return -code error "Data file exists."
  }
  if {[catch {esweep::size -obj $idorObj}]} {
    set ir [getProp $idorObj IR]
  } else {
    set ir $idorObj
  }
  # assemble meta data with no binary data in it
  set propList [::overlay::getProperty]
  foreach prop {IR GatedIR FR} {
    set i [lsearch $propList $prop]
    set propList [lreplace $propList $i $i]
  }
  set meta [list]
  foreach prop $propList {
    # ignore expanded microphone correction
    if {$prop eq {MicExp}} continue
    lappend meta $prop [getProp $idorObj $prop]
  }
  if {[catch {esweep::save -obj $ir -filename $fname -meta $meta} msg]} {
    return -code error $msg
  }
  return $meta
}

proc :l {{dirname {}} {replace no} {ignore no}} {
  if {[setup cget System,LastDir] eq {}} {
    set initdir [file join [pwd]]
  } else {
    set initdir [setup cget System,LastDir]
  }
  if {$dirname == {}} {
    if {[set dirname [tk_chooseDirectory \
      -initialdir $initdir \
      -mustexist yes]] == {}} {
      ::cmdbar::writeToCmdBar "Load cancelled" -bg yellow
      return 0
    }
  }
  if {![file isdirectory $dirname]} {
    if {$create} {
      file create $dirname
    } else {
      return -code error "Directory $dirname does not exist"
    }
  }
  setup configure System,LastDir $dirname
  if {[set idx [open [file join $dirname index] r]] eq {}} {
    return -code error "Couldn't open index file"
  }
  if {$replace} {
    ::overlay::delete all
  }
  for {set i 0} {[gets $idx line] >= 0} {incr i} {
    if {[catch {array set props $line}]} {
      close $idx
      return -code error {Index file possibly corrupted}
    }
    if {[catch {set fname [file join $dirname $props(Filename)]}]} {
      # property filename does not exist, try to use the "Name" property
      # (compat for older versions)
      if {[catch {set fname [file join $dirname $props(Name)]}]} {
        close $idx
        return -code error "No valid filename found in line: $i"
      }
    } else {
      # throw away the filename property
      array unset props Filename
    }
    # In older versions the path was absolute, leading to problems when moving a directory. 
    # That was a stupid idea, and now we're forced to check if it is a relative or absolute path. 
    if {[string index $fname 0] eq {.} || [file tail $fname] eq $fname} {
      # relative path
      set fname [file join $dirname $fname]
    }
    # the encoding conversion handles the automatic conversion done by esweep::load
    if {![file exists [encoding convertto utf-8 $fname]] && !$ignore} {
      close $idx
      return -code error "File $fname does not exist"
    }
    if {[catch {set props(IR) [esweep::load -filename $fname -meta meta]}]} {
      close $idx
      return -code error "Error loading file $fname"
    }
    if {[catch {set id [::overlay::add [array get props]]} msg]} {
      close $idx
      return -code error $msg
    }
    # get microphone compensation
    foreach {prop val} $meta {
      if {$prop eq {Mic}} {
        set props(Mic) $val
        set props(MicExp) [expandMicCompensation $val [esweep::samplerate -obj $props(IR)]]
        ::overlay::modify $id [list Mic $val MicExp $props(MicExp)]
        break
      }
    }
    # set the visibility expicitly, otherwise it may not be shown/hidden
    ::overlay::modify $id [list Show $props(Show)]

    array unset props
  }
  close $idx

  # wait for the last id to get displayed
  :wait #$id FR 

  # set the ::master variable to trigger some updates
  set ::master $::master
  # success
  ::cmdbar::writeToCmdBar "$dirname loaded" -bg green
  return 0
}

proc :export {{ID #1} {fname {}} {gated yes} {overwrite no}} {
  if {![regexp {(^#)([[:digit:]]+$)} $ID -> hash id]} {
      return -code error "Parameter not an ID: $ID"
  }
  if {$fname eq {}} {
    if {[set fname [tk_getSaveFile \
      -defaultextension {.ir} \
      -filetypes {
        {{ASCII files}  {.txt}  }
        {{Binary files} {.ir}   }
        {{All files}   *       }
      }
      ]] eq {}
    } {
      ::cmdbar::writeToCmdBar "Save cancelled" -bg yellow
      return 0
    } else {
      set overwrite yes
    }
  }
  if {$gated eq {yes}} {
    set ir [getProp $id IR]
    set FS [esweep::samplerate -obj $ir]
    lassign [lmap t $gate {expr {$t > 0.0 ? int(0.5+$t*$FS/1000.0) : 0}}] gate_start gate_stop
    if {$gate_stop >= [esweep::size -obj $ir]} {set gate_stop [expr {[esweep::size -obj $ir]-1}]}
    set gate_length [expr {$gate_stop-$gate_start}]
    set ir_gated [esweep::create -type [esweep::type -obj $ir] -size $gate_length -samplerate $FS]
    esweep::copy -dst ir_gated -src $ir -srcIdx $gate_start
  } else {
    set ir_gated [getProp $id IR]
  }

  if {[file extension $fname] ne {.txt}} {
    # write binary
    if {$gated eq {yes}} {
      writeFile $ir_gated $fname $overwrite
    } else {
      writeFile $id $fname $overwrite
    }
  } else {
    esweep:toAscii -filename $fname -obj $ir_gated
  }
}

proc :import {{fname {}}} {
  if {$fname eq {}} {
    if {[set fname [tk_getOpenFile \
      -defaultextension {.ir} \
      -filetypes {
        {{Binary files} {.ir}   }
        {{All files}   *       }
      }
      ]] eq {}
    } {
      ::cmdbar::writeToCmdBar "Load cancelled" -bg yellow
      return 0
    }
  }
  if {[catch {set ir [esweep::load -filename $fname -meta meta]}]} {
    return -code error "Error loading file $fname"
  }
  if {[esweep::type -obj $ir] ne {wave} && [esweep::type -obj $ir] ne {complex}} {
    return -code error "File wrong format"
  }
  # set properties by hand
  lappend props Name [expr {[set i [lsearch $meta Name]] >= 0 ? [lindex $meta $i+1] : {Master}}]
  lappend props Color [expr {[set i [lsearch $meta Color]] >= 0 ? [lindex $meta $i+1] : {Red}}]
  lappend props SigType [expr {[set i [lsearch $meta SigType]] >= 0 ? [lindex $meta $i+1] : {unknown}}]
  lappend props SweepRate [expr {[set i [lsearch $meta SweepRate]] >= 0 ? [lindex $meta $i+1] : -1}]
  lappend props Gate [expr {[set i [lsearch $meta Gate]] >= 0 ? [lindex $meta $i+1] : [setup cget Processing,Gate]}]
  lappend props Window [expr {[set i [lsearch $meta Window]] >= 0 ? [lindex $meta $i+1] : \
                             [list [setup cget Processing,Gate,Window,Type] [setup cget Processing,Gate,Window,Width]]}]
  lappend props Mode [expr {[set i [lsearch $meta Mode]] >= 0 ? [lindex $meta $i+1] : {single}}]
  lappend props OutLvl [expr {[set i [lsearch $meta OutLvl]] >= 0 ? [lindex $meta $i+1] : 1.0}]
  # always show
  lappend props Show {yes}
  lappend props IR $ir

  if {[::overlay::exists $::master]} {
    ::overlay::modify $::master $props
  } else {
    set ::master [::overlay::add $props]
  }
  ::overlay::select set $::master
  # copy master to buffer
  ::overlay::copy
}
