# Copyright 2016 Jochen Fabricius
# Distribution and modification is encouraged

################
# Key bindings #
################

proc bindings {win} {
  bind $win <Escape> [list ::audio::cancel]

  # program control commands
  bind $win <m> [list :measure default]

  # overlay controls
  bind $win <Double-d> [list apply {{} {
      global master
      set IDs [::overlay::select get]
      if {[lsearch $IDs $master] < 0} {
        ::overlay::yank
      } else {
        ::cmdbar::writeToCmdBar {Master curve may not be deleted} -bg red
      }
    }}]
  bind $win <Double-y> [list apply {{} {
      :copy
      }}]
  bind $win <p> [list apply {{} {
      :paste
    }}]
  # show selected overlays
  bind $win <S> [list apply {{} {
      foreach ID [::overlay::select get] {
        ::overlay::modify $ID [list Show {yes}]
        ::overlay::enable $ID
      }
    }}]
  bind $win <s> [bind $win <S>]
  #
  # hide selected overlays
  bind $win <H> [list apply {{} {
      global master
      foreach ID [::overlay::select get] {
        ::overlay::modify $ID [list Show {no}]
        ::overlay::disable $ID
      }
    }}]
  bind $win <h> [bind $win <H>]

  # change color
  bind $win <C> [list apply {{win} {
      foreach sel [::overlay::select get] {
        lappend IDs #$sel
      }
      :color $IDs
   }} $win]
   bind $win <c> [bind $win <C>]

   # rename
   bind $win <R> [list apply {{} {:rename}}]
   bind $win <r> [bind $win <R>]
   bind $win <F2> [bind $win <R>]

   # make selected overlay the master
   bind $win <M> [list apply {{} {
       set ::master [lindex [::overlay::select get] 0]
       # we need to update the harmonic distortion
       # we can't do it in the trace handler because it might give an error with the very first
       # measurement due to a race condition
       updateHD
       if {[::TkEsweepXYPlot::isVisible HD]} {::TkEsweepXYPlot::plot HD}
     }}]
   bind $win <Control-m> [bind $win <M>]
   
   # make selected curve the cursor master and set cursor 1 to it
   bind $win {1} [list apply {{} {
     if {[set ID [lindex [:select] 0]] eq {}} {
       set ID [:master]
     }
     :cursor trace 1 $ID
     set ::diff(Master) [isID $ID]
   }}]

   # make selected curve the cursor slave and set cursor 2 to it
   bind $win <Key-2> [list apply {{} {
     if {[set ID [lindex [:select] 0]] eq {}} {
       set ID [:master]
     }
     #:cursor trace 2 $ID
     set id [isID $ID]
     if {$id ne $::diff(Slave)} {
       # either set ...  
       set ::diff(Slave) $id
     } else {
       # or release slave
       :cursor trace 2 $ID
       set ::diff(Slave) {}
     }
   }}]

   bind $win <.> [list apply {{} {
       ::cmdbar::execCmdBar [::cmdbar::historyGet]
     }
   }]

   bind $win <F1> :help

   # overlay navigation
   bind $win <Up> [list apply {{} {if {[focus] ne [::overlay::win]} {::overlay::select prev}}}]
   bind $win <Down> [list apply {{} {if {[focus] ne [::overlay::win]} {::overlay::select next}}}]
   bind $win <Prior> [list apply {{} {for {set i 0} {i < 5} {incr i} {::overlay::select prev}}}]
   bind $win <Next> [list apply {{} {for {set i 0} {i < 5} {incr i} {::overlay::select next}}}]
   bind $win <Home> [list apply {{} {::overlay::select set 1}}]
   bind $win <End> [list apply {{} {::overlay::select end}}]
}

##############################
# Frequency response diagram #
##############################

proc createPlots {fr ir hd bd} {
  global master
  ::TkEsweepXYPlot::init $fr FR
  set ::TkEsweepXYPlot::plots(FR,Config,Title) {Frequency response}

  set ::TkEsweepXYPlot::plots(FR,Config,Font,Size) {-12}

  set ::TkEsweepXYPlot::plots(FR,Config,Margin,Top) 50
  set ::TkEsweepXYPlot::plots(FR,Config,Margin,Bottom) 80
  set ::TkEsweepXYPlot::plots(FR,Config,Margin,Left) 100
  set ::TkEsweepXYPlot::plots(FR,Config,Margin,Right) 100
  set ::TkEsweepXYPlot::plots(FR,Config,MouseHoverActive) {no}
  set ::TkEsweepXYPlot::plots(FR,Config,Legend,Align) center
  set ::TkEsweepXYPlot::plots(FR,Config,Cursors,1) {on}
  #set ::TkEsweepXYPlot::plots(FR,Config,Cursors,1,Trace) $::master
  set ::TkEsweepXYPlot::plots(FR,Config,Cursors,2) {on}
  trace add variable ::TkEsweepXYPlot::plots(FR,Config,Cursors,1,X) write [list apply {{v1 v2 op} {
      updateBD
    }}]


  set ::TkEsweepXYPlot::plots(FR,Config,X,Label) {f [Hz]}
  set ::TkEsweepXYPlot::plots(FR,Config,X,Precision) 5
  set ::TkEsweepXYPlot::plots(FR,Config,X,Autoscale) {off}
  if {[setup cget Display,FR,HRange] eq {auto}} {
    set ::TkEsweepXYPlot::plots(FR,Config,X,Min) 20
    set ::TkEsweepXYPlot::plots(FR,Config,X,Max) 20e3
  } else {
    set ::TkEsweepXYPlot::plots(FR,Config,X,Min) [lindex [setup cget Display,FR,HRange] 0]
    set ::TkEsweepXYPlot::plots(FR,Config,X,Max)  [lindex [setup cget Display,FR,HRange] 1]
  }

  set ::TkEsweepXYPlot::plots(FR,Config,X,Log) {yes}

  set ::TkEsweepXYPlot::plots(FR,Config,Y1,Label) {Mag [dB]}
  set ::TkEsweepXYPlot::plots(FR,Config,Y1,Label) "Magnitude\n\[dB SPL / [setup cget Display,Scaling,Reference] V\]"
  set ::TkEsweepXYPlot::plots(FR,Config,Y1,Precision) 3
  set ::TkEsweepXYPlot::plots(FR,Config,Y1,MStep) 5
  set ::TkEsweepXYPlot::plots(FR,Config,Y1,Autoscale) {on}
  set ::TkEsweepXYPlot::plots(FR,Config,Y1,Min) -[setup cget Display,FR,VRange]
  set ::TkEsweepXYPlot::plots(FR,Config,Y1,Max) 0
  set ::TkEsweepXYPlot::plots(FR,Config,Y1,Range) [setup cget Display,FR,VRange]
  set ::TkEsweepXYPlot::plots(FR,Config,Y1,Log) {no}

  if {0} {
  trace add variable ::TkEsweepXYPlot::plots(FR,Config,Cursors,1,Trace) write [list apply {{v1 v2 op} {
      upvar $v1 tr
      set ::diff(Master) $tr($v2)
    }}]
  trace add variable ::TkEsweepXYPlot::plots(FR,Config,Cursors,2,Trace) write [list apply {{v1 v2 op} {
      upvar $v1 tr
      set ::diff(Slave) $tr($v2)
    }}]
  }
  
  ::TkEsweepXYPlot::plot FR

  ::TkEsweepXYPlot::init $ir IR
  set ::TkEsweepXYPlot::plots(IR,Config,Title) {Impulse response}

  set ::TkEsweepXYPlot::plots(IR,Config,Font,Size) {-12}

  set ::TkEsweepXYPlot::plots(IR,Config,Margin,Top) 50
  set ::TkEsweepXYPlot::plots(IR,Config,Margin,Bottom) 80
  set ::TkEsweepXYPlot::plots(IR,Config,Margin,Left) 100
  set ::TkEsweepXYPlot::plots(IR,Config,Margin,Right) 100
  set ::TkEsweepXYPlot::plots(IR,Config,MouseHoverActive) {no}
  set ::TkEsweepXYPlot::plots(IR,Config,Legend,Align) center
  set ::TkEsweepXYPlot::plots(IR,Config,Cursors,1) {on}
  set ::TkEsweepXYPlot::plots(IR,Config,Cursors,1,Trace) $::master
  set ::TkEsweepXYPlot::plots(IR,Config,Cursors,2) {on}
  set ::TkEsweepXYPlot::plots(IR,Config,Cursors,2,Trace) $::master
  ::TkEsweepXYPlot::setCursorByValue IR 1 0
  ::TkEsweepXYPlot::setCursorByValue IR 2 100
  # wait for the cursors to be drawn before entering the display update process
  
  trace add execution ::TkEsweepXYPlot::setCursor leave [list apply {{cmdstr code result op} {
      if {[lindex $cmdstr 1] ne {IR}} {return}
      set x0 [::TkEsweepXYPlot::getCursor IR 1]
      set x1 [::TkEsweepXYPlot::getCursor IR 2]
      if {$x0 < $x1} {
        if {[setup cget Processing,Gate,Mode] eq {manual}} {
          setup configure Processing,Gate [list $x0 $x1]
        } else {
          ::cmdbar::writeToCmdBar {Gating in "auto" mode, gate not altered.} -bg yellow
        }
      } else {
        ::cmdbar::writeToCmdBar {Cursor 1 > Cursor 2. Can't update FR.} -bg yellow
      }
    }}]

  set ::TkEsweepXYPlot::plots(IR,Config,X,Label) {t [ms]}
  set ::TkEsweepXYPlot::plots(IR,Config,X,Precision) 3
  set ::TkEsweepXYPlot::plots(IR,Config,X,Step) 5
  set ::TkEsweepXYPlot::plots(IR,Config,X,Autoscale) {off}
  set ::TkEsweepXYPlot::plots(IR,Config,X,Min) 0
  set ::TkEsweepXYPlot::plots(IR,Config,X,Max) 100
  set ::TkEsweepXYPlot::plots(IR,Config,X,Log) {no}

  set ::TkEsweepXYPlot::plots(IR,Config,Y1,Label) {Amplitude [Pa]}
  set ::TkEsweepXYPlot::plots(IR,Config,Y1,Label) "Amplitude\n\[Pa / [setup cget Display,Scaling,Reference] V\]"
  set ::TkEsweepXYPlot::plots(IR,Config,Y1,Precision) 3
  set ::TkEsweepXYPlot::plots(IR,Config,Y1,MStep) 5
  set ::TkEsweepXYPlot::plots(IR,Config,Y1,Autoscale) {on}
  set ::TkEsweepXYPlot::plots(IR,Config,Y1,Log) {no}

  updateIR -1
  ::TkEsweepXYPlot::plot IR

  ::TkEsweepXYPlot::init $hd HD
  set ::TkEsweepXYPlot::plots(HD,Config,Title) {Harmonic distortion}

  set ::TkEsweepXYPlot::plots(HD,Config,Font,Size) {-12}

  set ::TkEsweepXYPlot::plots(HD,Config,Margin,Top) 50
  set ::TkEsweepXYPlot::plots(HD,Config,Margin,Bottom) 80
  set ::TkEsweepXYPlot::plots(HD,Config,Margin,Left) 100
  set ::TkEsweepXYPlot::plots(HD,Config,Margin,Right) 100
  set ::TkEsweepXYPlot::plots(HD,Config,MouseHoverActive) {no}
  set ::TkEsweepXYPlot::plots(HD,Config,Legend,Align) center
  set ::TkEsweepXYPlot::plots(HD,Config,Cursors,1) {on}
  set ::TkEsweepXYPlot::plots(HD,Config,Cursors,1,Trace) $::master
  set ::TkEsweepXYPlot::plots(HD,Config,Cursors,2) {off}

  set ::TkEsweepXYPlot::plots(HD,Config,X,Label) {f [Hz]}
  set ::TkEsweepXYPlot::plots(HD,Config,X,Precision) 6
  set ::TkEsweepXYPlot::plots(HD,Config,X,Autoscale) {off}
  if {[setup cget Display,HD,HRange] eq {auto}} {
    set ::TkEsweepXYPlot::plots(HD,Config,X,Min) 20
    set ::TkEsweepXYPlot::plots(HD,Config,X,Max) 20e3
  } else {
    set ::TkEsweepXYPlot::plots(HD,Config,X,Min) [lindex [setup cget Display,HD,HRange] 0]
    set ::TkEsweepXYPlot::plots(HD,Config,X,Max)  [lindex [setup cget Display,HD,HRange] 1]
  }

  set ::TkEsweepXYPlot::plots(HD,Config,X,Log) {yes}

  set ::TkEsweepXYPlot::plots(HD,Config,Y1,Label) {Mag [dB]}
  set ::TkEsweepXYPlot::plots(HD,Config,Y1,Label) "Magnitude\n\[dB SPL / [setup cget Display,Scaling,Reference] V\]"
  set ::TkEsweepXYPlot::plots(HD,Config,Y1,Precision) 3
  set ::TkEsweepXYPlot::plots(HD,Config,Y1,Step) 8
  set ::TkEsweepXYPlot::plots(HD,Config,Y1,MStep) 5
  set ::TkEsweepXYPlot::plots(HD,Config,Y1,Autoscale) {on}
  set ::TkEsweepXYPlot::plots(HD,Config,Y1,Min) -[setup cget Display,HD,VRange]
  set ::TkEsweepXYPlot::plots(HD,Config,Y1,Max) 0
  set ::TkEsweepXYPlot::plots(HD,Config,Y1,Range) [setup cget Display,HD,VRange]
  set ::TkEsweepXYPlot::plots(HD,Config,Y1,Log) {no}

  updateHD -1
  catch {::TkEsweepXYPlot::plot HD}

  ::TkEsweepXYPlot::init $bd BD
  set ::TkEsweepXYPlot::plots(BD,Config,Title) {Burst decay}

  set ::TkEsweepXYPlot::plots(BD,Config,Font,Size) {-12}

  set ::TkEsweepXYPlot::plots(BD,Config,Margin,Top) 50
  set ::TkEsweepXYPlot::plots(BD,Config,Margin,Bottom) 80
  set ::TkEsweepXYPlot::plots(BD,Config,Margin,Left) 100
  set ::TkEsweepXYPlot::plots(BD,Config,Margin,Right) 100
  set ::TkEsweepXYPlot::plots(BD,Config,MouseHoverActive) {no}
  set ::TkEsweepXYPlot::plots(BD,Config,Legend,Align) center
  set ::TkEsweepXYPlot::plots(BD,Config,Cursors,1) {off}
  set ::TkEsweepXYPlot::plots(BD,Config,Cursors,2) {off}

  set ::TkEsweepXYPlot::plots(BD,Config,X,Label) {Periods}
  set ::TkEsweepXYPlot::plots(BD,Config,X,Precision) 3
  set ::TkEsweepXYPlot::plots(BD,Config,X,Step) 5
  set ::TkEsweepXYPlot::plots(BD,Config,X,Autoscale) {off}
  set ::TkEsweepXYPlot::plots(BD,Config,X,Min) 0
  if {[setup cget Display,BD,HRange] eq {auto}} {
    set ::TkEsweepXYPlot::plots(BD,Config,X,Max) 20
  } else {
    set ::TkEsweepXYPlot::plots(BD,Config,X,Max)  [setup cget Display,BD,HRange]
  }

  set ::TkEsweepXYPlot::plots(BD,Config,X,Min) 0
  set ::TkEsweepXYPlot::plots(BD,Config,X,Max) 20
  set ::TkEsweepXYPlot::plots(BD,Config,X,Log) {no}

  set ::TkEsweepXYPlot::plots(BD,Config,Y1,Label) "Amplitude\[dBPa re Peak]"
  set ::TkEsweepXYPlot::plots(BD,Config,Y1,Precision) 3
  set ::TkEsweepXYPlot::plots(BD,Config,Y1,MStep) 5
  set ::TkEsweepXYPlot::plots(BD,Config,Y1,Autoscale) {off}
  set ::TkEsweepXYPlot::plots(BD,Config,Y1,Min) -[setup cget Display,BD,VRange]
  set ::TkEsweepXYPlot::plots(BD,Config,Y1,Max) 0
  set ::TkEsweepXYPlot::plots(BD,Config,Y1,Log) {no}

  #updateBD
  ::TkEsweepXYPlot::plot BD
}

proc :cursor {subcmd cursor args} {
  switch -- $subcmd {
    set {
      ::TkEsweepXYPlot::setCursorByValue FR $cursor [lindex $args 0]
      ::TkEsweepXYPlot::plot FR
    }
    trace {
      switch -- [set arg [lindex $args 0]] {
        get {
          return -code ok #$::TkEsweepXYPlot::plots(FR,Config,Cursors,$cursor,Trace)
        }
        default {
          set id [isID $arg]
          set ::TkEsweepXYPlot::plots(FR,Config,Cursors,$cursor,Trace) $id
          set ::TkEsweepXYPlot::plots(FR,Config,Cursors,$cursor,Color) [getProp $id Color]
          set ::TkEsweepXYPlot::plots(FR,Config,Cursors,$cursor,Prefix) "$arg: "
          ::TkEsweepXYPlot::plot FR
        }
      }
    }
    default {
      return -code error "Unknown subcommand $subcmd"
    }
  }
}

proc updateDiff {v1 v2 op} {
  upvar $v1 diff

  DEBUGPRINT "$diff(Master) <-> $diff(Slave)"
  if {($diff(Master) == {}) && ($diff(Slave) == {})} {
    # unset cursor trace
    set ::TkEsweepXYPlot::plots(FR,Config,Cursors,2,Trace) {}
  } elseif {($diff(Master) != {}) && ($diff(Slave) == {})} {
    # set cursor to master
    set ::TkEsweepXYPlot::plots(FR,Config,Cursors,2,Trace) $diff(Master)
  } elseif {($diff(Master) == {}) && ($diff(Slave) != {})} {
    # set cursor to slave
    set ::TkEsweepXYPlot::plots(FR,Config,Cursors,2,Trace) $diff(Slave)
  } else {
    # create diff
    set master [getProp $diff(Master) FR]
    set slave [getProp $diff(Slave) FR]
    # check for correct mapping of both frequency responses
    if { ([llength $master] > 0) \
      && ([llength $master] == [llength $slave]) \
      && ([lindex $master 0] == [lindex $slave 0]) \
      && ([lindex $master end-2] == [lindex $slave end-2])} {
      foreach {x1 a1 p1} $master {x2 a2 p2} $slave {
        lappend fr $x1 [expr {$a1-$a2}] [expr {$p1-$p2}]
      }
      # add the diff response, but don't show it
      if {[::TkEsweepXYPlot::exists FR $diff(ID)]} {
        ::TkEsweepXYPlot::configTrace FR $diff(ID) -trace $fr -state {off} -name {diff}
      } else {
        ::TkEsweepXYPlot::addTrace FR $diff(ID) {diff} $fr
        ::TkEsweepXYPlot::configTrace FR $diff(ID) -state {off}
      }
      set ::TkEsweepXYPlot::plots(FR,Config,Cursors,2,Trace) $diff(ID)
      set ::TkEsweepXYPlot::plots(FR,Config,Cursors,2,Prefix) {dy }
      set ::TkEsweepXYPlot::plots(FR,Config,Cursors,2,Color) [getProp $diff(Slave) Color]
      ::TkEsweepXYPlot::plot FR
    }
    # else: leave cursor as it is
  }
}

proc setVisibility {id} {
  if {$id > 0} {
    if {[getProp $id Show] eq {no}} {
      DEBUGPRINT "Disabling ID #$id"
      ::overlay::disable $id
      catch {::TkEsweepXYPlot::configTrace FR $id -state off}
      catch {::TkEsweepXYPlot::configTrace IR $id -state off}
      catch {::TkEsweepXYPlot::configTrace BD $id -state off}
    } else {
      DEBUGPRINT "Enabling ID #$id"
      ::overlay::enable $id
      catch {::TkEsweepXYPlot::configTrace FR $id -state on}
      catch {::TkEsweepXYPlot::configTrace IR $id -state on}
      catch {::TkEsweepXYPlot::configTrace BD $id -state on}
    }
  }
}

proc updateHD {{id {}}} {
  DEBUGPRINT "Updating ID #$id called by [info level 1]"
  #if {![::TkEsweepXYPlot::isVisible HD]} {return}
  if {$id eq {}} {set id $::master}
  if {$id < 0} {return}
  if {![::overlay::exists $id]} {return}
  if {[set sweeprate [getProp $id SweepRate]] < 0} {
    return -code error {HD: impulse response not generated with a logarithmic sweep.}
  }

  set name [getProp $id Name]
  set ir [getProp $id IR]
  set gated_ir [getProp $id GatedIR]
  set color [getProp $id Color]
  set mode [getProp $id Mode]
  set gate [getProp $id Gate]
  set window [getProp $id Window]
  set mic [getProp $id MicExp]

  if {[setup cget Display,Scaling] eq {relative} || $mode eq {dual}} {
    set sc [setup cget Display,Scaling,Reference]
    set ::TkEsweepXYPlot::plots(HD,Config,Y1,Label) "Magnitude \[dB SPL / [setup cget Display,Scaling,Reference] V\]"
  } elseif {[setup cget Display,Scaling] eq {absolute}} {
    set ::TkEsweepXYPlot::plots(HD,Config,Y1,Label) "Magnitude \[dB SPL\]"
    set sc [getProp $id OutLvl]
  } else {
    return -code error "Unknown scaling method \"[setup cget Display,Scaling]\""
  }
  set ::TkEsweepXYPlot::plots(HD,Config,Y1,Range) [setup cget Display,HD,VRange]

  set sc [expr {$sc/2e-5}]
  if {[catch {set hd [calcHD $ir $sc $gate [setup cget Processing,FFT,Size] $sweeprate $window [setup cget Processing,HD,List] \
      -weighting [setup cget Processing,HD,Weighting] \
      -shift     [setup cget Processing,HD,Shifting] \
      -mode      [setup cget Processing,HD,Mode] \
      -smoothing [setup cget Processing,FFT,Smoothing]]} msg]} {
    return -code error $msg
  }

  set colors [list \
    red \
    blue \
    green \
    purple \
    black \
    brown \
    darkblue \
    darkgreen \
    red4 \
  ]

  ::TkEsweepXYPlot::removeTrace HD all
  # plot frequency response
  ::TkEsweepXYPlot::addTrace HD 1 ${name} [set fund [::TkEsweepXYPlot::getTrace FR $id]]
  ::TkEsweepXYPlot::configTrace HD 1 -color [lindex $colors 0] -width 2

  foreach {k fr} $hd {
    # get the log spaced frequency response with the same mapping as the fundamental
    set coords [getLogSpacedFr $fr $fund $mic]
    ::TkEsweepXYPlot::addTrace HD $k "${name}: HD$k" $coords
    ::TkEsweepXYPlot::configTrace HD $k -color [lindex $colors [expr {$k%[llength $colors]}]] -width 2
  }

  if {[set shift [setup cget Processing,HD,FRShift]] > 0} {
    foreach {f a p} [::TkEsweepXYPlot::getTrace FR $id] {
      lappend fr_shifted $f [expr {$a-$shift}] $p
    }
    set id [::TkEsweepXYPlot::addTrace HD end "${name}: Limit" $fr_shifted]
    ::TkEsweepXYPlot::configTrace HD $id -color [lindex $colors 0] -width 2
  }
  # apply x-axis autoscaling if necessary
  if {[setup cget Display,HD,HRange] eq {auto}} {
    set ::TkEsweepXYPlot::plots(HD,Config,X,Max) 20e3
    # autoscaling may not be possible yet, so catch it
    set ::TkEsweepXYPlot::plots(HD,Config,X,Autoscale) {on}
    catch {::TkEsweepXYPlot::autoscale HD X}
    set ::TkEsweepXYPlot::plots(HD,Config,X,Autoscale) {off}
    set ::TkEsweepXYPlot::plots(HD,Config,X,Min) 20
  }

  ::TkEsweepXYPlot::plot HD
  DEBUGPRINT {Done}
}

proc updateIR {{id {}}} {
  #if {![::TkEsweepXYPlot::isVisible IR]} {return}
  if {$id < 0} {return}
  if {$id eq {}} {set id $::master}
  if {![::overlay::exists $id]} {return}
  set name [getProp $id Name]
  set ir [esweep::clone -src [getProp $id IR]]
  set color [getProp $id Color]
  set mode [getProp $id Mode]
  set gate [getProp $id Gate]
  lassign $gate g0 g1
  # apply display scaling
  if {[setup cget Display,Scaling] eq {relative} || $mode eq {dual}} {
    set sc [setup cget Display,Scaling,Reference]
  } elseif {[setup cget Display,Scaling] eq {absolute}} {
    set sc [getProp $id OutLvl]
  } else {
    return -code error "Unknown scaling method \"[setup cget Display,Scaling]\""
  }
  esweep::toWave -obj ir
  esweep::expr ir * $sc
  # draw the impulse response
  if {[::TkEsweepXYPlot::exists IR $id]} {
    ::TkEsweepXYPlot::configTrace IR $id -trace $ir -color $color -width 0 -state {on} -name $name
  } else {
    ::TkEsweepXYPlot::addTrace IR $id $name $ir
    ::TkEsweepXYPlot::configTrace IR $id -color $color -width 0
  }
  if {$id == $::master} {
    #set ::TkEsweepXYPlot::plots(IR,Config,X,Min) $g0
    #set ::TkEsweepXYPlot::plots(IR,Config,X,Max) $g1
    ::TkEsweepXYPlot::setCursorByValue IR 1 $g0
    ::TkEsweepXYPlot::setCursorByValue IR 2 $g1
  }
  #::TkEsweepXYPlot::plot IR
}

proc updateGatedIR {{id {}}} {
  if {$id < 0} {return}
  if {$id eq {}} {set id $::master}
  set ir [getProp $id IR]
  set gate [getProp $id Gate]
  set window [getProp $id Window]

  set ir_gated [calcGatedIR $ir $gate $window]

  ::overlay::modify $id [list GatedIR $ir_gated]
}

# update the log-spaced frequency response if the IR has changed
proc updateFR {{id {}}} {
  DEBUGPRINT "Updating ID #$id called by [info level 1]"
  if {$id eq {}} {set id $::master}

  if {$id >= 0} {
    set name [getProp $id Name]
    set ir [getProp $id GatedIR]
    set color [getProp $id Color]
    set mode [getProp $id Mode]
    set gate [getProp $id Gate]
    set window [getProp $id Window]
    set mic [getProp $id MicExp]
    
    # apply display scaling
    if {[setup cget Display,Scaling] eq {relative} || $mode eq {dual}} {
      set ::TkEsweepXYPlot::plots(FR,Config,Y1,Label) "Magnitude \[dB SPL / [setup cget Display,Scaling,Reference] V\]"
      set sc [setup cget Display,Scaling,Reference]
    } elseif {[setup cget Display,Scaling] eq {absolute}} {
      set ::TkEsweepXYPlot::plots(FR,Config,Y1,Label) "Magnitude \[dB SPL\]"
      set sc [getProp $id OutLvl]
    } else {
      return -code error "Unknown scaling method \"[setup cget Display,Scaling]\""
    }
    # scale to dB SPL
    set sc [expr {$sc/2e-5}]
    set fr_method [setup cget Processing,FR,Method]
    DEBUGPRINT "IR Length: [esweep::size -obj $ir], FFT size: [setup cget Processing,FFT,Size]"
    if {[catch {set fr [calc$fr_method $ir $sc [setup cget Processing,FFT,Size] [setup cget Processing,FFT,Smoothing]]} msg]} {
      return -code error $msg
    }
    set fr [getLogSpacedFr $fr \
      [list \
        $::TkEsweepXYPlot::plots(FR,Config,X,Min) \
        [expr {[esweep::samplerate -obj $fr]/2.1}] \
        [setup cget Display,Bins] \
        ] \
        $mic]
    # store this for future redraws  
    ::overlay::modify $id [list FR $fr]
    # draw the frequency response
    if {[getProp $id Show] ne {yes}} {
      set state off
    } else {
      set state on
    }
    if {[::TkEsweepXYPlot::exists FR $id]} {
      if {[getProp $id Show] ne {yes}} {
        set state off
      } else {
        set state on
      }
      ::TkEsweepXYPlot::configTrace FR $id -trace $fr -color $color -width 2 -state $state -name $name
    } else {
      ::TkEsweepXYPlot::addTrace FR $id $name $fr
      ::TkEsweepXYPlot::configTrace FR $id -color $color -width 2 -state $state
    }
    # apply x-axis autoscaling if necessary
    if {[setup cget Display,FR,HRange] eq {auto}} {
      set ::TkEsweepXYPlot::plots(FR,Config,X,Max) 20e3
      # autoscaling may not be possible yet, so catch it
      set ::TkEsweepXYPlot::plots(FR,Config,X,Autoscale) {on}
      catch {::TkEsweepXYPlot::autoscale FR X}
      set ::TkEsweepXYPlot::plots(FR,Config,X,Autoscale) {off}
      set ::TkEsweepXYPlot::plots(FR,Config,X,Min) 20
    }
  }
}

# add the frequency response to the plot or modify its name, color etc.
proc showFR {id} {
  if {$id >= 0} {
    set name [getProp $id Name]
    set fr [getProp $id FR]
    set color [getProp $id Color]

    # draw the frequency response
    if {[::TkEsweepXYPlot::exists FR $id]} {
      ::TkEsweepXYPlot::configTrace FR $id -trace $fr -color $color -width 2 -state {on} -name $name
    } else {
      ::TkEsweepXYPlot::addTrace FR $id $name $fr
      ::TkEsweepXYPlot::configTrace FR $id -color $color -width 2
    }
  }
}

proc updateBD {{cycles 5} {params {}}} {
  if {![::TkEsweepXYPlot::isVisible BD]} {return}
  # workaround for strange cmdbar behaviour
  if {$cycles eq {}} {set cycles 5}
  set ID $::master
  if {![::overlay::exists $ID]} {return}

  if {[llength $params] == 0} {
    set freqs [::TkEsweepXYPlot::getCursor FR 1]
  } else {
    foreach f $params {
      lappend freqs $f
    }
  }
  # use a 5 cycle blackman shaped cosine
  #set cycles 5
  set P 20

  ::TkEsweepXYPlot::removeTrace BD all

  # setup the gated IR for later calculations
  set ir [getProp $ID IR]
  set gate [getProp $ID Gate]
  set window [getProp $ID Window]
  set color [getProp $ID Color]
  set Fs [esweep::samplerate -obj $ir]
  lassign [lmap t $gate {expr {$t > 0.0 ? int(0.5+$t*$Fs/1000.0) : 0}}] gate_start gate_stop
  set ir_gated [esweep::create -type [esweep::type -obj $ir] -size [expr {$gate_stop-$gate_start}] -samplerate $Fs]
  esweep::copy -dst ir_gated -src $ir -srcIdx $gate_start
  esweep::toWave -obj ir_gated
  esweep::window -obj ir_gated \
    -rightwin [lindex $window 0] -rightwidth [lindex $window 1]
  #esweep::fft -inplace -obj ir_gated

  # use the first frequency as a visual reference
  lassign $freqs f
  if {$f <= 0 || $f > [expr {$Fs/2.0}]} {
    return -code error "Invalid frequency: $f"
  }
  # create a complex burst, this is already the analytic signal
  set burst [esweep::create -type complex -samplerate $Fs -size [expr {int(0.5+double($cycles)/$f*$Fs)}]]
  esweep::generate sine -obj burst -frequency $f
  esweep::window -obj burst -leftwin blackman -leftwidth 50 -rightwin blackman -rightwidth 50
  esweep::abs -obj burst
  esweep::toWave -obj burst
  esweep::clip -obj burst -clip 1e-6
  esweep::lg -obj burst
  esweep::expr burst * 20
  esweep::expr burst - [esweep::max -obj $burst]
  # fake a period scale by secretly adjusting the samplerate
  esweep::setSamplerate -obj burst -samplerate [expr {int(0.5+1000*$Fs/$f)}]
  set id [::TkEsweepXYPlot::addTrace BD end {Blackman Burst} [esweep::get -obj $burst -from 0 -to [expr {int(0.5+double($P)/$f*$Fs)}]]]
  ::TkEsweepXYPlot::configTrace BD $id -color {lightgrey} -width 2 -dash {-}


  set k 0
  foreach f $freqs {
    if {$f <= 0 || $f > [expr {$Fs/2.0}]} {
      return -code error "Invalid frequency: $f"
    }

    set burst [esweep::create -type complex -samplerate $Fs -size [expr {int(0.5+double($cycles)/$f*$Fs)}]]
    esweep::generate sine -obj burst -frequency $f
    esweep::window -obj burst -leftwin blackman -leftwidth 50 -rightwin blackman -rightwidth 50
    set maxPeak [esweep::maxPos -obj $burst]
    esweep::convolve -signal burst -filter $ir_gated
    esweep::abs -obj burst
    esweep::toWave -obj burst
    esweep::clip -obj burst -clip 1e-6
    esweep::lg -obj burst
    esweep::expr burst * 20
    esweep::expr burst - [esweep::max -obj $burst]
    # fake a period scale by secretly adjusting the samplerate
    esweep::setSamplerate -obj burst -samplerate [expr {int(0.5+1000*$Fs/$f)}]
    switch -- [setup cget Processing,BD,Align] {
      {peaks} {
        esweep::rotate -obj burst -samples [expr {abs($maxPeak-[esweep::maxPos -obj $burst])}] -dir {left}
      }
      default {}
    }

    set id [::TkEsweepXYPlot::addTrace BD end [format {%s @ %.2f Hz} [getProp $ID Name] $f] [esweep::get -obj $burst -from 0 -to [expr {int(0.5+double($P)/$f*$Fs)}]]]
    ::TkEsweepXYPlot::configTrace BD $id -color $color -width 2
    incr k
  }
  ::TkEsweepXYPlot::plot BD
}



proc createLevelMeter {w} {
  # the level meter is a frame within a frame
  # while the outer frame keeps constant in size
  # the inner frame can be adjusted in height by calling
  # setLevelMeter
  frame $w.left  -relief raised -background red4 -bd 2
  frame $w.right  -relief raised -background red4 -bd 2

  ttk::label $w.l -text {L}
  ttk::label $w.r -text {R}

  place $w.l -relx 0 -rely 0.0 -relwidth 0.5 -anchor nw
  place $w.r -relx 0.5 -rely 0.0 -relwidth 0.5 -anchor nw
  place $w.left -relx 0 -rely 1.0 -relwidth 0.5 -relheight 0.0 -anchor sw
  place $w.right -relx 0.5 -rely 1.0 -relwidth 0.5 -relheight 0.0 -anchor sw

  # create traces on the maxR and maxL variables
  ::audio::strace add [::audio::map Left],Level,RMS write [list setLevelMeter $w.left]
  ::audio::strace add [::audio::map Right],Level,RMS write [list setLevelMeter $w.right]
}

proc setLevelMeter {w name1 name2 op} {
  upvar $name1 var
  # Color code:
  # $level < -20 dB: red (much too quiet)
  # -20 dB < $level < -6 dB: yellow (too quiet)
  # -6 dB < $level < -1 dB: green (OK)
  # -1 dB < $level: red (overload)
  set level [expr {20*log10($var($name2) < 1e-5 ? 1e-5 : $var($name2))}]
  if {$level < -64} {
    set color red4
  } elseif {$level < -48} {
    set color red
  } elseif {$level < -36} {
    set color yellow
  } elseif {$level < -1} {
    set color green
  } else {
    set color red
  }
  # configure level meter
  $w configure -background $color
  place configure $w -relheight [expr {($level+100)/100.0}]
}

proc appWindows {} {
  ####################
  # The command line #
  ####################
  ::cmdbar::init .cmdBar
  set ::cmdbar::options(Debug) 0

  set t [frame .top]
  #########################
  # The level meter frame #
  #########################
  set i [frame $t.input]

  ###################
  # The level meter #
  ###################
  frame $i.level -width 30 -bd 2 -relief sunken
  createLevelMeter $i.level
  pack $i.level -side right -fill y -expand no

  ##################################
  # The frequency response diagram #
  ##################################
  set f [frame $t.fr]
  set o [frame $t.o]
  pack $f -expand yes -fill both -side left
  pack $o -expand no -fill both -side left
  pack $i -fill y -side left
  pack $t -side top -fill both -expand yes
  pack .cmdBar -side top -fill x

  ##########################################
  # Impulse response diagram in a toplevel #
  ##########################################
  set ir [toplevel .ir -width 800 -height 600]
  update idletasks
  wm transient $ir .
  wm withdraw $ir
  wm protocol $ir WM_DELETE_WINDOW [list :hide IR]
  # zoom in
  bind $ir <z> [list apply {{} {
    if {[set x0 [::TkEsweepXYPlot::getCursor IR 1]] eq {}} {set x0 0}
    if {[set x1 [::TkEsweepXYPlot::getCursor IR 2]] eq {}} {set x1 0}
    if {$x0 > $x1} {
      set t $x0
      set x0 $x1
      set x1 $t
      ::TkEsweepXYPlot::setCursorByValue IR 1 $x0
      ::TkEsweepXYPlot::setCursorByValue IR 2 $x1
    }
    set ::TkEsweepXYPlot::plots(IR,Config,X,Min) $x0
    set ::TkEsweepXYPlot::plots(IR,Config,X,Max) $x1
    ::TkEsweepXYPlot::plot IR
  }}]
  bind $ir <Z> [list apply {{} {
    if {[set x0 [::TkEsweepXYPlot::getCursor IR 1]] eq {}} {set x0 0}
    if {[set x1 [::TkEsweepXYPlot::getCursor IR 2]] eq {}} {set x1 0}
    if {$x0 > $x1} {
      set t $x0
      set x0 $x1
      set x1 $t
      ::TkEsweepXYPlot::setCursorByValue IR 1 $x0
      ::TkEsweepXYPlot::setCursorByValue IR 2 $x1
    }
    set x0 $::TkEsweepXYPlot::plots(IR,Config,X,Min)
    set x1 $::TkEsweepXYPlot::plots(IR,Config,X,Max)
    set xm [expr {($x0+$x1)/2.0}]
    set dx [expr {double($x1-$x0)}]
    set ::TkEsweepXYPlot::plots(IR,Config,X,Min) [expr {$xm-$dx}]
    set ::TkEsweepXYPlot::plots(IR,Config,X,Max) [expr {$xm+$dx}]
    if {$::TkEsweepXYPlot::plots(IR,Config,X,Min) < 0} {set ::TkEsweepXYPlot::plots(IR,Config,X,Min) 0.0}
    ::TkEsweepXYPlot::plot IR
  }}]
  bind $ir <Home> [list apply {{} {
    set ::TkEsweepXYPlot::plots(IR,Config,X,Min) 0.0
    ::TkEsweepXYPlot::setCursorByValue IR 1 0.0
    ::TkEsweepXYPlot::plot IR
  }}]
  bind $ir <Control-Home> [list apply {{} {
    ::TkEsweepXYPlot::setCursorByValue IR 1 $::TkEsweepXYPlot::plots(IR,Config,X,Min)
    ::TkEsweepXYPlot::plot IR
  }}]
  bind $ir <End> [list apply {{} {
    global master
    if {![catch {set ir [getProp $master IR]}]} {
      set ::TkEsweepXYPlot::plots(IR,Config,X,Max) [expr {(1000.0*[esweep::size -obj $ir])/[esweep::samplerate -obj $ir]}]
      ::TkEsweepXYPlot::setCursorByValue IR 2 $::TkEsweepXYPlot::plots(IR,Config,X,Max)
      ::TkEsweepXYPlot::plot IR
    }
  }}]
  bind $ir <Control-End> [list apply {{} {
    ::TkEsweepXYPlot::setCursorByValue IR 2 $::TkEsweepXYPlot::plots(IR,Config,X,Max)
    ::TkEsweepXYPlot::plot IR
  }}]
  bind $ir <Left> [list apply {{} {
    set ::TkEsweepXYPlot::plots(IR,Config,X,Min) [expr {$::TkEsweepXYPlot::plots(IR,Config,X,Min)-($::TkEsweepXYPlot::plots(IR,Config,X,Max)-$::TkEsweepXYPlot::plots(IR,Config,X,Min))/10.0}]
    set ::TkEsweepXYPlot::plots(IR,Config,X,Max) [expr {$::TkEsweepXYPlot::plots(IR,Config,X,Max)-($::TkEsweepXYPlot::plots(IR,Config,X,Max)-$::TkEsweepXYPlot::plots(IR,Config,X,Min))/10.0}]
    ::TkEsweepXYPlot::plot IR
  }}]
  bind $ir <Right> [list apply {{} {
    set ::TkEsweepXYPlot::plots(IR,Config,X,Min) [expr {$::TkEsweepXYPlot::plots(IR,Config,X,Min)+($::TkEsweepXYPlot::plots(IR,Config,X,Max)-$::TkEsweepXYPlot::plots(IR,Config,X,Min))/10.0}]
    set ::TkEsweepXYPlot::plots(IR,Config,X,Max) [expr {$::TkEsweepXYPlot::plots(IR,Config,X,Max)+($::TkEsweepXYPlot::plots(IR,Config,X,Max)-$::TkEsweepXYPlot::plots(IR,Config,X,Min))/10.0}]
    ::TkEsweepXYPlot::plot IR
  }}]

  set hd [toplevel .hd -width 800 -height 600]
  update idletasks
  wm transient $hd .
  wm withdraw $hd
  wm protocol $hd WM_DELETE_WINDOW [list :hide HD]

  set bd [toplevel .bd -width 800 -height 600]
  update idletasks
  wm transient $hd .
  wm withdraw $bd
  wm protocol $bd WM_DELETE_WINDOW [list :hide BD]

  # create the plot
  createPlots $f $ir $hd $bd
  update

  initOverlays $o

  ##############################
  # application title and icon #
  ##############################
  updateAppTitle FR
  updateAppTitle IR
  updateAppTitle HD

  # constrain window size and aspect ratio
  #wm aspect . 4 3 16 9
  wm minsize . 800 600

  focus $t

  wm protocol . WM_DELETE_WINDOW [list :q]
}

proc updateAppTitle {win} {
  switch -- $win {
    FR {
      set title {FR }
      append title  {| Mode: } [setup cget Measure,Mode] \
                    { | FFT: } [expr {[setup cget Processing,FFT,Size] > 0 ? [setup cget Processing,FFT,Size] : {auto}}] \
                    { | Gate: } [setup cget Processing,Gate,Mode] \
                    { | FS: } [expr {[setup cget Audio,Samplerate]/1000}] { kHz}
      wm title . $title
    }
    HD {
      set title {HD }
      append title {| Mode: } [setup cget Processing,HD,Mode] { | Weighting: } [setup cget Processing,HD,Weighting]
      wm title .hd $title
    }
    IR {
      set title {IR}
      wm title .ir $title
    }
  }
}

