# Copyright 2016 Jochen Fabricius
# Distribution and modification is encouraged

# common math functions for the cmdbar

proc :expr {ID1 op IDorExpr} {
  if {[catch {isID $ID1} id1]} {
    return -code error $id1
  }
  set fr1 [getProp $id1 FR]
  if {[catch {isID $IDorExpr} id2]} {
    # test for expression
    # note that there are no curly braces around the argument for expr. This ensures that the expression will be evaluated
    if {[catch {set a [expr [lindex $IDorExpr 0]]}]} {
      return -code error "Second operand neither an ID nor an expression: $IDorExpr"
    }
    if {[llength $IDorExpr] == 2} {
      # maybe this is a complex number in polar notation
      if {[catch {set p [expr [lindex $IDorExpr 1]]}]} {
        return -code error "Second operand is polar format, but phase component is not usable: $IDorExpr"
      }
    } else {
      # set phase to zero
      set p 0.0
    }
    # fake a frequency response with same mapping as fr1. This makes the calculation below way easier
    set a [expr {20*log10($a)}]
    foreach {f _ _} $fr1 {
      lappend fr2 $f $a $p
    }
  } else {
    set fr2 [getProp $id2 FR]
  }

  if { [llength $fr1] != [llength $fr2] \
    || [lindex $fr1 0] != [lindex $fr2 0] \
    || [lindex $fr1 end-2] != [lindex $fr2 end-2]} {
      return -code error "$ID1 and $ID2 have different mappings"
  }

  switch -- $op {
    {+} {
      # adding to polar numbers requires a conversion to complex format and backwards
      foreach {x1 a1 p1} $fr1 {x2 a2 p2} $fr2 {
        set ar [expr {pow(10, $a1/20.0)}]; set r1 [expr {$ar*cos($p1)}]; set i1 [expr {$ar*sin($p1)}]; 
        set ar [expr {pow(10, $a2/20.0)}]; set r2 [expr {$ar*cos($p2)}]; set i2 [expr {$ar*sin($p2)}]; 
        set r [expr {$r1+$r2}]; set i [expr {$i1+$i2}]; 

        lappend fr $x1 [expr {10*log10($r**2+$i**2)}] [expr {atan2($i, $r)}]
      }
    }
    {-} {
      foreach {x1 a1 p1} $fr1 {x2 a2 p2} $fr2 {
        set ar [expr {pow(10, $a1/20.0)}]; set r1 [expr {$ar*cos($p1)}]; set i1 [expr {$ar*sin($p1)}]; 
        set ar [expr {pow(10, $a2/20.0)}]; set r2 [expr {$ar*cos($p2)}]; set i2 [expr {$ar*sin($p2)}]; 
        set r [expr {$r1-$r2}]; set i [expr {$i1-$i2}]; 

        lappend fr $x1 [expr {10*log10($r**2+$i**2)}] [expr {atan2($i, $r)}]
      }
    }
    {*} {
      foreach {x1 a1 p1} $fr1 {x2 a2 p2} $fr2 {
        lappend fr $x1 [expr {$a1+$a2}] [expr {$p1+$p2}]
      }
    }
    {/} {
      foreach {x1 a1 p1} $fr1 {x2 a2 p2} $fr2 {
        lappend fr $x1 [expr {$a1-$a2}] [expr {$p1-$p2}]
      }
    }
    default {
      return -code error "Unknown operator $op"
    }
  }
  ::overlay::modify $id1 [list FR $fr]

  return $ID1
}

